import logging
import multiprocessing
import socket

from qw.client import Client
from qw.exception import AlreadyStartedException, NotStartedException
from qw.worker import Worker


class Manager(object):
    __slots__ = ["workers", "client", "num_workers", "log", "name", "target"]

    def __init__(self, target, host="localhost", port=6379, db=0, num_workers=None, name=None):
        self.workers = []
        self.num_workers = num_workers or multiprocessing.cpu_count()
        self.client = Client(host=host, port=port, db=db)
        self.log = logging.getLogger("qw.manager")
        self.name = name or socket.gethostname()
        self.target = target

    def start(self):
        self.log.info("starting", extra={"process_name": self.name})
        if self.workers:
            raise AlreadyStartedException("Workers Already Started")

        self.log.info("starting %s workers", self.num_workers, extra={"process_name": self.name})
        for _ in xrange(self.num_workers):
            worker = Worker(self.client, self.target, manager_name=self.name)
            worker.start()
            self.workers.append(worker)
        self.log.info("registering %s", self.name, extra={"process_name": self.name})
        self.client.register_manager(self.name)

    def join(self):
        self.log.debug("joining workers", extra={"process_name": self.name})
        if not self.workers:
            raise NotStartedException("Workers Do Not Exist")

        # wait for them all to stop
        for worker in self.workers:
            worker.join()

        # make sure to clear out workers list
        self.workers = []

    def stop(self):
        self.log.info("stopping", extra={"process_name": self.name})
        if not self.workers:
            raise NotStartedException("Workers Do Not Exist")

        self.log.info("deregistering %s", self.name, extra={"process_name": self.name})
        self.client.deregister_manager(self.name)

        self.log.info("shutting down %s workers", len(self.workers), extra={"process_name": self.name})
        # trigger them all to shutdown
        for worker in self.workers:
            worker.shutdown()

        self.join()
        self.log.info("stopped", extra={"process_name": self.name})



import logging

FORMAT = "%(asctime)s - %(name)s - %(process_name)s - %(levelname)s - %(message)s"
formatter = logging.Formatter(FORMAT)

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)

manager_logger = logging.getLogger("qw.manager")
manager_logger.addHandler(stream_handler)

worker_logger = logging.getLogger("qw.worker")
worker_logger.addHandler(stream_handler)



import logging
import multiprocessing
import traceback
import socket
import os

from qw.utils import dynamic_import


class Worker(multiprocessing.Process):
    __slots__ = [
        "client", "exit", "log", "timeout", "manager_name", "target"
    ]

    def __init__(self, client, target, manager_name=None, timeout=10):
        super(Worker, self).__init__()
        self.client = client
        self.manager_name = manager_name or socket.gethostname()
        self.exit = multiprocessing.Event()
        self.timeout = timeout
        self.log = logging.getLogger("qw.worker")
        self.target = target

    @property
    def name(self):
        return "%s.%s" % (self.manager_name, os.getpid())

    @property
    def job_queue(self):
        return "%s:jobs" % (self.name, )

    def _register(self):
        self.log.info(
            "registering worker %s under '%s:workers'" % (self.name, self.manager_name),
            extra={"process_name": self.name}
        )
        self.client.register_worker(self.manager_name, self.name)

    def _deregister(self):
        self.log.info(
            "deregistering worker %s from '%s:workers'" % (self.name, self.manager_name),
            extra={"process_name": self.name}
        )
        self.client.deregister_worker(self.manager_name, self.name)

    def _run(self):
        # try to fetch a previously unfinished job
        # otherwise try to fetch from the main job pool
        self.log.debug("polling for jobs", extra={"process_name": self.name})
        job_id, job_data = self.client.fetch_next_job(self.manager_name, self.name, timeout=self.timeout)
        if not job_id or not job_data:
            return

        self.log.debug(
            "processing job id (%s) data (%r)" % (job_id, job_data), extra={"process_name": self.name}
        )
        if job_data:
            self.target(job_id, job_data)

        self.log.debug("removing job id (%s)" % (job_id), extra={"process_name": self.name})
        self.client.finish_job(job_id, self.name)

    def run(self):
        self.log.info("starting", extra={"process_name": self.name})
        if isinstance(self.target, basestring):
            self.target = dynamic_import(self.target)
        self._register()
        while not self.exit.is_set():
            try:
                self._run()
            except KeyboardInterrupt:
                self.log.error("encountered a KeyboardInterrupt", extra={"process_name": self.name})
                break
            except Exception, e:
                self.log.error("encountered an error (%r)" % (e, ), extra={"process_name": self.name})
                traceback.print_exc()

        self._deregister()
        self.log.info("stopping", extra={"process_name": self.name})

    def shutdown(self):
        self.log.info("shutdown signal received", extra={"process_name": self.name})
        self.exit.set()



import redis

import uuid


class Client(redis.StrictRedis):
    ALL_MANAGERS = "all:managers"
    ALL_JOBS = "all:jobs"
    MANAGER_WORKERS = "%s:workers"
    MANAGER_JOBS = "%s:jobs"
    WORKER_JOBS = "%s:jobs"
    JOB_KEY = "job:%s"

    def __init__(self, host="localhost", port=6379, db=0):
        super(Client, self).__init__(host=host, port=port, db=db)

    def register_manager(self, name):
        self.sadd(self.ALL_MANAGERS, name)

    def deregister_manager(self, name)    :
        self.srem(self.ALL_MANAGERS, name)

    def register_worker(self, manager, name):
        self.sadd(self.MANAGER_WORKERS % (manager, ), name)

    def deregister_worker(self, manager, name):
        self.srem(self.MANAGER_WORKERS % (manager, ), name)

    def queue_job(self, job_data, manager=None, worker=None):
        job_id = uuid.uuid4()
        self.hmset(self.JOB_KEY % (job_id, ), job_data)
        if manager is not None:
            self.lpush(self.MANAGER_JOBS % (manager, ), job_id)
        elif worker is not None:
            self.lpush(self.WORKER_JOBS % (worker, ), job_id)
        else:
            self.lpush(self.ALL_JOBS, job_id)

        return job_id

    def fetch_next_job(self, manager, worker, timeout=10):
        # try to fetch in this order
        #  1) any jobs already assigned but not finished by the worker
        #  2) any jobs assigned specifically to the manager
        #  3) try to grab a job from the pool of all jobs
        job_id = (
            self.lpop(self.WORKER_JOBS % (worker, )) or
            self.rpoplpush(self.MANAGER_JOBS % (manager, ), self.WORKER_JOBS % (worker, )) or
            self.brpoplpush(self.ALL_JOBS, self.WORKER_JOBS % (worker, ), timeout=timeout)
        )

        job_data = None
        if job_id is not None:
            job_data = self.hgetall(self.JOB_KEY % (job_id, ))

        return (job_id, job_data)

    def finish_job(self, job_id, worker_name):
        self.delete(self.JOB_KEY % (job_id, ))
        self.lrem(self.WORKER_JOBS % (worker_name, ), 1, job_id)

    def get_all_managers(self):
        return self.smembers(self.ALL_MANAGERS)

    def get_manager_workers(self, manager_name):
        return self.smembers(self.MANAGER_WORKERS % (manager_name, ))

    def get_worker_pending_jobs(self, worker_name):
        for job_id in self.lrange(self.WORKER_JOBS % (worker_name, ), 0, -1):
            yield job_id

    def get_manager_queued_jobs(self, manager_name):
        for job_id in self.lrange(self.MANAGER_JOBS % (manager_name, ), 0, -1):
            yield job_id

    def get_all_queued_jobs(self):
        for job_id in self.lrange(self.ALL_JOBS, 0, -1):
            yield (None, job_id)

        for manager in self.get_all_managers():
            for job_id in self.get_manager_queued_jobs(manager):
                yield (manager, job_id)

    def get_all_pending_jobs(self):
        for manager in self.get_all_managers():
            for worker in self.get_manager_workers(manager):
                for job_id in self.get_worker_pending_jobs(worker):
                    yield (worker, job_id)



#!/usr/bin/env python

from setuptools import setup, find_packages

from qw import __version__

setup(
    name="qw",
    version=__version__,
    description="Python Distributed Redis Queue Workers",
    author="Brett Langdon",
    author_email="brett@blangdon.com",
    url="https://github.com/brettlangdon/qw",
    packages=find_packages(),
    license="MIT",
    scripts=["bin/qw-manager", "bin/qw-client"],
    install_requires=[
        "redis==2.10.3",
        "docopt==0.6.2",
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: MIT License",
        "Topic :: Utilities",
    ]
)



def dynamic_import(name):
    module, _, function = name.rpartition(".")
    mod = __import__(module, fromlist=[function])
    components = name.split(".")
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod
